import logging
from .decoders import decoders
from .unprefix import unprefix


logger = logging.getLogger(__name__)


def env_get(d, key, default, decoders=decoders, required=None):
    """
    Look up ``key`` in ``d`` and decode it, or return ``default``.
    """
    if required is None:
        required = isinstance(default, type)

    try:
        value = d[key]
    except KeyError:
        if required:
            raise
        return default

    dt = (default if default is None or isinstance(default, type)
          else type(default))

    for decoder in decoders:
        if (decoder.decodes_to_type(dt) and
            decoder.decodes_from_value(value)
        ):
            try:
                return decoder.decode(value)
            except Exception as e:
                logger.error("%s couldn't convert %s: %s: %s",
                             decoder.__class__.__name__, key,
                             e.__class__.__name__, e)
                raise

    raise ValueError("no suitable env decoder for {}".format(key))





    return (_envb(d, key, default) if isinstance(default, bool) else
            _envi(d, key, default) if isinstance(default, int) else
            _envj(d, key, default) if isinstance(default, (list, dict)) else
            _envs(d, key, default))




def _envs(d, key, default):
    return d.get(key, default)


def _envb(d, key, default):
    value = d.get(key)
    if value is None:
        return default
    if isinstance(value, bool):
        return value
    if isinstance(value, string_types):
        v = value.lower()
        if v in ['1', 'on', 't', 'true', 'y', 'yes']:
            return True
        if v in ['0', 'off', 'f', 'false', 'n', 'no']:
            return False
    logger.warning('%s bool value is invalid (%r)', key, value)
    return default


def _envi(d, key, default):
    value = d.get(key)
    if value is None:
        return default
    if isinstance(value, int):
        return value
    if isinstance(value, string_types):
        try:
            return int(value)
        except ValueError:
            pass
    logger.warning('%s int value is invalid (%r)', key, value)
    return default


def _envj(d, key, default):
    value = d.get(key)
    if value is None:
        return default
    if isinstance(value, string_types):
        try:
            return json.loads(value)
        except json.JSONDecodeError as e:
            logger.warning('%s JSON value is %r, %s: %s', key, value,
                           e.__class__.__name__, e)
    return default



import abc
import json


class EnvDecoder:
    """
    Base class for type-specific decoders.
    """

    # This metaclass prevents this class from being instantiated, and
    # subclasses must override any abstract methods (``from_string``) before
    # they can be instantiated as well.
    __metaclass__ = abc.ABCMeta

    # resulting type(s) of decoding
    to_types = ()

    @property
    def from_types(self):
        """
        Types the decoder handles as input.
        Most decoders handle their output types plus strings.
        """
        return self.to_types + (str,)

    def decodes_to_type(self, type):
        """
        Test if decoder can decode to type
        """
        return type in self.to_types

    def decodes_from_value(self, value):
        """
        Test if decoder can decode from value
        """
        return isinstance(value, self.from_types)

    def decode(self, value):
        """
        Decode the given value. Raises ValueError if type(value) isn't handled.
        """
        if isinstance(value, self.to_types):
            return value
        if isinstance(value, str):
            return self.from_string(value)
        raise ValueError("don't know {}".format(type(value)))  # pragma: no cover

    @abc.abstractmethod
    def from_string(self, s):
        """
        Decode from string ``s``, or raise an exception on failure.
        Must be overridden by subclass, and don't call super()
        """
        raise AssertionError("EnvDecoder.from_string shouldn't be called!")  # pragma: no cover


class EnvBool(EnvDecoder):
    """
    Decoder for boolean values. Handles various string representations of
    True/False.
    """

    to_types = (bool,)

    def from_string(self, s):
        s = s.lower()
        if s in ['1', 'on', 't', 'true', 'y', 'yes']:
            return True
        if s in ['0', 'off', 'f', 'false', 'n', 'no']:
            return False
        raise ValueError("can't interpret string as bool")


class EnvStr(EnvDecoder):
    """
    Decoder for string values. ``os.environ`` values should always be strings,
    so this is easy.
    """

    to_types = (str,)

    def from_string(self, s):
        # never actually called, because EnvDecoder.decode short circuits since
        # the input type matches the output type.
        return s  # pragma: no cover


class EnvInt(EnvDecoder):
    """
    Decoder for base 10 integer values.
    """

    to_types = (int,)

    def from_string(self, s):
        return int(s)


class EnvList(EnvDecoder):
    """
    Decoder for JSON-encoded lists.
    """

    to_types = (list,)

    def decodes_from_value(self, value):
        return isinstance(value, self.to_types) or (
            isinstance(value, str) and
            value.lstrip().startswith('[')
        )

    def from_string(self, s):
        return json.loads(s)


class EnvDict(EnvDecoder):
    """
    Decoder for JSON-encoded dicts.
    """

    to_types = (dict,)

    def decodes_from_value(self, value):
        return isinstance(value, self.to_types) or (
            isinstance(value, str) and
            value.lstrip().startswith('{')
        )

    def from_string(self, s):
        return json.loads(s)


class EnvNone(EnvDecoder):
    """
    Decode anything as a pass-through.
    """

    def decodes_to_type(self, type):
        return type is None

    def decodes_from_value(self, value):
        return True

    def decode(self, value):
        return value

    def from_string(self, s):
        raise AssertionError("EnvNone.from_string shouldn't be called!")  # pragma: no cover


decoders = [EnvBool(), EnvInt(), EnvList(), EnvDict(), EnvStr(), EnvNone()]


__all__ = ['env']



def unprefix(prefix, d, all=False):
    """
    Returns a new dict by removing ``prefix`` from keys.
    If ``all`` is ``False`` (default) then drops keys without the prefix,
    otherwise keeping them.
    """
    d1 = dict(d) if all else {}
    d1.update((k[len(prefix):], v) for k, v in d.items() if k.startswith(prefix))
    return d1



from setuptools import setup, find_packages
from codecs import open
import os

here = os.path.dirname(__file__)
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='bienvenue',
    version='3.0.0',
    description='Read config from environment for 12 factor apps',
    long_description=long_description,
    url='https://github.com/scampersand/bienvenue',
    author='Aron Griffis',
    author_email='aron@scampersand.com',
    license='MIT',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='django environment environ env'.split(),
    packages=find_packages(exclude=['tests']),
    install_requires=[],
)



from functools import partial
import os
from .env import env_get
from .unprefix import unprefix


def make_env_reader(d=os.environ, prefix=None, all=False, **kwargs):
    """
    Make a function that accesses a dict-like object, removing key prefixes and
    converting string values to types. Defaults to ``os.environ`` with no
    prefix.
    """
    if prefix:
        d = unprefix(prefix, d, all=all)
    return partial(env_get, d, **kwargs)
